import pickle
import logging
from random import shuffle, seed

from pesten.pesten import Pesten, card
from pesten.agent import Agent


logger = logging.getLogger(__name__)


def test_game():
    """Plays a game. Players just try playing cards until one of them wins"""
    cards = [card(suit, value) for suit in range(4) for value in range(13)]
    shuffle(cards)
    game = Pesten(4, 8, cards)
    curr = 0
    turn_count = 0
    while not game.has_won:
        choose = 0
        while True:
            new_curr = game.play_turn(choose)
            if choose < 0:
                break
            if new_curr != curr:
                curr = new_curr
                break
            choose += 1
            if choose >= len(game.curr_hand):
                choose = -1
        turn_count += 1
    logger.info(f"{turn_count=}")


def test_with_jokers():
    game = Pesten(2,2, [77,77,77,77,77,77,77,77,77,77,30,0,], {77: 'draw_card-5', 78: 'draw_card-5'})
    ai = Agent(99)

    current_player = 0
    chooses = [1, -1, 1, 1, 2, -1, 2, -1, 2, -1, 0, 1, -1, 0, 0]
    i = 0
    while i < len(chooses) or current_player == 1:
        logger.info(f"{i}: {current_player} - {game.current_hand()} {game.draw_count=}")
        if current_player == 0:
            assert game.check(chooses[i])
            current_player = game.play_turn(chooses[i])
            i += 1
        else:
            current_player = game.play_turn(ai.generate_choose(game))
    else:
        assert current_player == 0
        logger.info(f"{i}: {current_player} - {game.current_hand()} {game.draw_count=}")
    assert game.has_won


def test_reshuffle_on_empty_draw_stack():
    game = Pesten(2, 3, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], {})
    while len(game.current_hand()) > 1:
        game.play_turn(0)
    while len(game.draw_stack) > 0:
        game.play_turn(-1)
    assert len(game.play_stack) > 1
    assert len(game.draw_stack) == 0
    assert game.play_stack == [12, 11, 10, 9, 8]
    game.play_turn(-1)
    assert game.play_stack == [8]
    assert game.draw_stack == [12, 9, 11]


def test_reshuffle_on_empty_draw_stack_odd():
    game = Pesten(2, 3, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], {})
    while len(game.current_hand()) > 1:
        game.play_turn(0)
    while len(game.draw_stack) > 0:
        game.play_turn(-1)
    assert len(game.play_stack) > 1
    assert len(game.draw_stack) == 0
    assert game.play_stack == [12, 11, 10, 9, 8]
    game.play_turn(0)
    game.play_turn(-1)
    assert game.play_stack == [7]
    assert game.draw_stack == [8, 9, 10, 11]

def test_shuffle():
    game = Pesten(2, 2, [1,2,3,4,5,6,7,8,9], {})
    game.shuffle()


def test_played_game():
    with open("tests/testgame.pickle", 'rb') as file:
        game, _, chooses, _ = pickle.load(file)

    reconstructed_game = Pesten(4, 8, game.init_cards, {
        9: 'change_suit',
        0: 'draw_card-2',
        5: 'another_turn',
        6: 'skip_turn',
        12: 'reverse_order',
    })
    for choose in chooses:
        assert not reconstructed_game.has_won
        reconstructed_game.play_turn(choose)
    assert reconstructed_game.has_won


import logging

from pesten.pesten import Pesten, CannotDraw, EndWithSpecialCard


logger = logging.getLogger(__name__)


class AgentError(Exception):
    ...


class Agent:
    def __init__(self, player_index):
        self.player_index = player_index

    
    def get_possible_chooses(self, game: Pesten):
        card_count = len(game.curr_hand)
        possible_choosese = []
        for possible_choose in range(card_count):
            try:
                if game.check(possible_choose):
                    possible_choosese.append(possible_choose)
            except EndWithSpecialCard:
                continue
        return possible_choosese
    

    def generate_choose(self, game: Pesten):
        possible_choosese = self.get_possible_chooses(game)
        if possible_choosese:
            choose = possible_choosese[0]
        else:
            game.assert_can_draw()
            choose = -1
        
        if game.asking_suit:
            choose = 0
        elif game.draw_count > 0:
            choose = -1

        # Make AI counter if it has draw_card cards
        if "draw_card" in game.rules.get((game.play_stack[-1] % 13), ""):
            for possible_choose in possible_choosese:
                if "draw_card" in game.resolve_rule(possible_choose):
                    choose = possible_choose
        return choose

    
    def play_turn(self, game: Pesten):
        assert game.current_player == self.player_index
        if game.asking_suit:
            choose = 0
        elif game.drawing:
            choose = -1
        try:
            choose = self.generate_choose(game)
            topcard = game.play_stack[-1]
            index_next_player = game.play_turn(choose)
            if not game.has_won and index_next_player == self.player_index and topcard == game.play_stack[-1]:
                # If the turn stays the same and the topcard the same then the choose was not good
                raise AgentError("Wrong choose generated")
        except CannotDraw as e:
            raise AgentError from e
        except AgentError as e:
            raise e
        except Exception as e:
            raise AgentError("AI got an error", e)



import asyncio
import logging
from typing import Protocol
import json

from pesten.pesten import Pesten, CannotDraw, EndWithSpecialCard, card_object
from pesten.agent import Agent, AgentError

logger = logging.getLogger(__name__)


class ConnectionDisconnect(Exception):
    ...


class ClosingConnection(Exception):
    """This exception will prevent a NullConnection to get stuck in its gameloop inside Lobby.connect"""
    ...


class Connection(Protocol):
    async def accept(self): ...
    async def close(self): ...
    async def send_json(self, data): ...
    async def receive_text(self) -> str: ...


class NullConnection:
    async def accept(self): ...

    async def close(self): ...

    async def send_json(self, data): ...

    async def receive_text(self) -> str:
        raise ClosingConnection("Null connection closing")
        

class AIConnection():
    def __init__(self, game: Pesten, player_index, delay = 1): # Resolve player_index automatically
        self.game = game
        self.agent = Agent(player_index)
        self.event = asyncio.Event()
        self.delay = delay
        self.messages = []
        self.exit = False

    async def accept(self):
        ...
    
    async def close(self):
        self.event.set() # makes it raise exception in receive_text method to stop loop
        self.exit = True
    
    async def send_json(self, data: dict):
        # This function can trigger the event when it detects that its this AI its turn
        # logger.debug(f"Received for {self.agent.player_index} : \n{json.dumps(data, indent=2)}")
        if 'error' in data:
            return
        if self.game.current_player == self.agent.player_index or self.game.has_won:
            # if len(self.messages) < 3:
            #     self.messages.append(data)
            # else:
            #     if all([msg["current_player"] == data["current_player"] for msg in self.messages]):
            #         raise Exception("AI is getting stuck")
            #     self.messages = []
            self.event.set()

    
    async def receive_text(self) -> str:
        await self.event.wait() # Would be nice to get rid of this for not using create_tasks
        self.event.clear()
        if self.exit:
            raise ClosingConnection("Closing AI after exit")
        choose = self.agent.generate_choose(self.game)
        await asyncio.sleep(self.delay)
        return choose


class Player:
    # Structure holding player information 
    def __init__(self, name, connection: Connection):
        self.name = name
        self.connection = connection


class Lobby:
    def __init__(self, game: Pesten) -> None:
        self.game = game
        self.started = False
        self.capacity = game.player_count
        self.players: list[Player] = [] # List corresponds with players in pesten game
        self.chooses = []
        self.run = True

    async def connect(self, new_player: Player):
        name = new_player.name
        # Creates a gameloop for a connection
        if player := self.get_player_by_name(name):
            index = self.players.index(player)
            self.players[index] = new_player # Replacing the player object
        elif self.started:
            raise Exception("Lobby is full")
        else:
            self.players.append(new_player)
        if len(self.players) == self.capacity:
            self.started = True
        await new_player.connection.accept() # Accept as late as possible
        connection = new_player.connection
        await self.update_boards(message=f"{name} joined the game")
        # break-statements only make sure that the current connection stops
        while self.run:
            try:
                choose = await connection.receive_text()
                await self.play_choose(new_player, choose)
                self.run = not self.game.has_won # Stop all connections if game was won
            except CannotDraw as e:
                await new_player.connection.send_json({"error": "Cannot draw, you have to play a card"})
            except EndWithSpecialCard as e:
                logger.error("Player tried to end with special card")
                await new_player.connection.send_json({"error": "Cannot end with special card"})
            except ClosingConnection as e:
                break
            except ConnectionDisconnect as e:
                break

    async def update_boards(self, message=""):
        for player_id, player in enumerate(self.players):
            try:
                send_coro = player.connection.send_json({
                    "topcard": card_object(self.game.play_stack[-1]),
                    "previous_topcard": card_object(self.game.play_stack[-2]) if len(self.game.play_stack) > 1 else None,
                    "can_draw": bool(self.game.draw_stack),
                    "choose_suit": self.game.asking_suit,
                    "draw_count": self.game.draw_count,
                    "current_player": self.players[self.game.current_player].name,
                    "otherPlayers": {
                        self.players[i].name
                        if i < len(self.players)else "": 
                        len(self.game.hands[i])
                        if i < len(self.players) else 0
                        for i in range(self.capacity)
                    },
                    "hand": [card_object(card) for card in self.game.hands[player_id]],
                    "message": message
                })
                logger.debug(f"Updating {self.players[self.game.current_player].name}'s board")
                await send_coro
            except IndexError:
                # the amount of players should be at least the same as the current player index
                # Fix by having a NullConnection for every inital player
                continue

    def get_player_by_name(self, name: str) -> Player:
        try:
            player = next(filter(lambda p: p.name == name, self.players))
        except StopIteration as e:
            return None
        return player
        
    async def play_choose(self, player: Player, choose):
        # player = self.get_player_by_name(name)
        name = player.name
        if not self.started:
            await player.connection.send_json({"error": "Game not started"})
            return
        if self.game.current_player != self.players.index(player):
            await player.connection.send_json({"error": "Not your turn"})
            return
        try:
            choose = int(choose)
        except ValueError:
            await player.connection.send_json({"error": "Invalid choose"})
            return
        log_count_before = len(self.game.logs)
        self.chooses.append(choose)
        self.game.play_turn(choose)
        if len(self.game.logs) != log_count_before:
            message = self.game.logs[-1][1]
        else:
            message = ""
        if self.game.has_won:
            await self.update_boards(f"{name} has won the game!")
        else:
            await self.update_boards(message=message)


import logging
from typing import Literal


logger = logging.getLogger(__name__)

SUITS = ["hearts", "diamonds", "spades", "clubs"]
VALUES = ["2", "3", "4", "5", "6", "7", "8", "9", "10", "jack", "queen", "king", "ace"]
BLACK_JOKER = 77
RED_JOKER = 78


def card(suit, value):
    # suit: 0-4, value 0-13
    return suit*13 + value


def card_string(c):
    if c >= 52:
        value = 'mirror'
    else:
        value = VALUES[c % 13]
    return SUITS[(c // 13) % 4] + " " + value


def card_object(card):
    if card == RED_JOKER:
        return {"suit": 'red', "value": 'joker'}
    if card == BLACK_JOKER:
        return {"suit": 'black', "value": 'joker'}
    suit, value = card_string(card).split(' ')
    return {"suit": suit, "value": value}


class CannotDraw(Exception):
    ...


class EndWithSpecialCard(Exception):
    ...


class Pesten:
    def __init__(self, player_count: int, hand_count, cards: list, rules: dict = {}) -> None:
        self.init_cards = list(cards)
        self.player_count = player_count
        self.current_player = 0
        self.draw_stack = cards
        self.play_stack = [cards.pop()]
        self.hands = [[] for _ in range(player_count)]
        self.curr_hand = self.hands[self.current_player]
        self.reverse = False
        self.rules = rules
        self.change_suit_state: Literal["not asked", "asking", "asked"] = "not_asked"
        self.drawing = False
        self.draw_count = 0
        self.has_won = False
        self.asking_suit = False
        self.enable_logging = False
        self.logs = []
        for _ in range(hand_count):
            for hand in self.hands:
                hand.append(self.draw_stack.pop())


    def shuffle(self):
        new_stack = []
        while len(self.draw_stack) > 1:
            new_stack.append(self.draw_stack.pop(-1))
            new_stack.append(self.draw_stack.pop(0))
        if len(self.draw_stack) == 1:
            new_stack.append(self.draw_stack.pop())
        assert not self.draw_stack
        self.draw_stack = new_stack
    
    
    def assert_can_draw(self):
        if len(self.draw_stack) + len(self.play_stack) <= 1:
            raise CannotDraw("Not enough cards on the board to draw. Please play a card")

    
    def log(self, message):
        if not self.enable_logging:
            return
        data = [self.current_player, message]
        if len(self.logs) > 0 and data == self.logs[-1]:
            return
        self.logs.append(data)


    def draw(self):
        self.assert_can_draw()
        if not self.draw_stack and len(self.play_stack) > 1:
            top_card = self.play_stack.pop()
            while self.play_stack:
                card = self.play_stack.pop()
                if card >= 52 and card != BLACK_JOKER and card != RED_JOKER:
                    # These were added when choosing suit. Don't put back in draw stack
                    continue
                self.draw_stack.append(card)
            for _ in range(100):
                self.shuffle()
            self.play_stack.append(top_card)
        self.curr_hand.append(self.draw_stack.pop())


    def check(self, choose):
        played_card = self.curr_hand[choose]
        top_card = self.play_stack[-1]
        if played_card == BLACK_JOKER or played_card == RED_JOKER or top_card == BLACK_JOKER or top_card == RED_JOKER:
            return True
        suit_top_card = (top_card // 13) % 4 # There should only be 52 cards with 51 being the highest
        can_play = played_card // 13 == suit_top_card or played_card % 13 == top_card % 13
        if can_play:
            self.log(f"Choose {choose}")
        # Not allowed to end with special card
        #TODO: Make this configurable
        #TODO: Joker can by-pass this
        is_special = (played_card % 13) in self.rules
        if is_special and len(self.current_hand()) == 1:
            self.log("Can't end with rule card")
            raise EndWithSpecialCard()
        return can_play


    def play(self, choose):
        self.play_stack.append(self.curr_hand.pop(choose))
        if not self.curr_hand:
            self.log("The game was won!")
            self.has_won = True


    def next(self):
        if self.has_won:
            return
        if not self.reverse:
            self.current_player += 1
        else:
            self.current_player -= 1
        self.current_player += self.player_count # Make sure it is a positive number
        self.current_player %= self.player_count # before modding
        self.curr_hand = self.hands[self.current_player]


    def resolve_rule(self, choose):
        value_choose = self.curr_hand[choose]
        if value_choose != BLACK_JOKER and value_choose != RED_JOKER:
            value_choose = value_choose % 13
        return self.rules.get(value_choose, "")
        

    def _play_turn(self, choose) -> int:
        # Returns index player who's next turn will come from.
        # If choose is negative it will draw a card

        # I explicitly return in all cases to be explicit about the flow
        if self.has_won:
            return int(self.current_player)


        if self.asking_suit:
            if choose >= len(SUITS):
                # Asking again
                return self.current_player
            self.chosen_suit = choose
            value_card = self.play_stack[-1] % 13
            suit_card = 52 + choose * 13 + value_card
            self.log(f"Suit choosen: {SUITS[choose]}")
            self.play_stack.append(suit_card) # Will be removed on reshuffling deck
            self.next()
            self.asking_suit = False
            return self.current_player


        if self.draw_count > 0:
            if choose < 0:
                self.log(f"Has to draw {self.draw_count}")
                for _ in range(self.draw_count):
                    try:
                        self.draw()
                    except CannotDraw:
                        self.log(f"Not enough cards to draw {self.draw_count}")
                        break
                self.draw_count = 0
                return self.current_player
            rule = self.resolve_rule(choose)
            if 'draw_card' not in rule or not self.check(choose):
                return self.current_player
            self.log("Countered with another draw card")
            # Continue as normal because I'm sure it will enter the draw_card if-block later

        if choose < 0:
            self.log("Drawing card")
            self.draw()
            self.next()
            return int(self.current_player)

        if choose < len(self.curr_hand):
            rule = self.resolve_rule(choose)
            if rule == 'another_turn':
                if self.check(choose):
                    self.log(f"Another turn {card_string(self.curr_hand[choose])}")
                    self.play(choose)
                return self.current_player
            
            if rule == 'skip_turn':
                if self.check(choose):
                    self.log(f'skip turn with {card_string(self.curr_hand[choose])}')
                    self.play(choose)
                    self.next()
                    self.next()
                return self.current_player
            
            if rule == 'reverse_order':
                if self.check(choose):
                    self.log(f'reverse order with {card_string(self.curr_hand[choose])}')
                    self.play(choose)
                    self.reverse = not self.reverse
                    self.next()
                return self.current_player

            if rule and 'draw_card' in rule:
                if self.check(choose):
                    _, count = rule.split("-")
                    self.draw_count += int(count)
                    self.log(f'draw card with {card_string(self.curr_hand[choose])}. counter: {self.draw_count}')
                    self.play(choose)
                    # self.drawing = True
                    self.next()
                return self.current_player

            if rule == 'change_suit':
                if self.check(choose):
                    self.log(f'changed suit with {card_string(self.curr_hand[choose])}')
                    self.play(choose)
                    self.asking_suit = True
                return self.current_player
                
            # default play
            if self.check(choose):
                self.log(f"played {card_string(self.curr_hand[choose])}")
                self.play(choose)       
                self.next()
            return self.current_player
        return self.current_player
    
    def play_turn(self, choose) -> int:
        self.enable_logging = True # Makes sure logs are only added by this function
        current_player = self._play_turn(choose)
        self.enable_logging = False
        return current_player

    def current_hand(self):
        return self.hands[self.current_player]



import logging
import asyncio
import random

import pytest


logger = logging.getLogger(__name__)


def test_agent_super_simple():
    from pesten.pesten import Pesten, card
    from pesten.agent import Agent

    game = Pesten(2, 1, [
        card(0, 0),
        card(0, 0),
        card(0, 0),
        card(0, 0),
    ])
    agent = Agent(0) 
    agent.play_turn(game)
    assert game.has_won


def test_agent_agent_error():
    from pesten.pesten import Pesten, card
    from pesten.agent import Agent, AgentError

    game = Pesten(2, 1, [
        card(0, 4),
        card(1, 5),
        card(2, 6),
        card(3, 7),
    ])
    agents = [Agent(i) for i in range(2)]
    with pytest.raises(AgentError):
        while not game.has_won: 
            agents[game.current_player].play_turn(game)


def test_agent_full_game():
    from pesten.pesten import Pesten, card
    from pesten.agent import Agent

    cards = [card(suit, value) for suit in range(4) for value in range(13)]
    random.seed(1)
    random.shuffle(cards)
    game = Pesten(2, 1, cards)
    agents = [Agent(i) for i in range(2)]
    turn_counter = 0
    while not game.has_won:
        choose = agents[turn_counter % 2].generate_choose(game)
        game.play_turn(choose)
        turn_counter += 1
    logger.info(f"Total amount of turns {turn_counter}")


def test_agent_wont_end_with_special_card():
    from pesten.pesten import Pesten, card
    from pesten.agent import Agent

    game = Pesten(2, 1, [0,0,0,0,0,0,0,0,0,0], {0: 'some_rule'})
    agent = Agent(0)
    game.play_turn(agent.generate_choose(game))


@pytest.mark.asyncio
async def test_two_ais_playing():
    from pesten.pesten import Pesten, card
    from pesten.lobby import Lobby, Player, AIConnection
    cards = [card(suit, value) for suit in range(4) for value in range(13)]
    random.seed(1)
    random.shuffle(cards)
    game = Pesten(2, 8, cards)
    lobby = Lobby(game)
    await asyncio.gather(
        lobby.connect(Player('player1', AIConnection(game, 0, delay=0))),
        lobby.connect(Player('player2', AIConnection(game, 1, delay=0)))
    )


@pytest.mark.asyncio
async def test_three_ais_playing():
    from pesten.pesten import Pesten, card
    from pesten.lobby import Lobby, Player, AIConnection
    cards = [card(suit, value) for suit in range(4) for value in range(13)]
    random.seed(1)
    random.shuffle(cards)
    game = Pesten(3, 8, cards)
    lobby = Lobby(game)
    await asyncio.gather(
        lobby.connect(Player('player0', AIConnection(game, 0, delay=0))),
        lobby.connect(Player('player1', AIConnection(game, 1, delay=0))),
        lobby.connect(Player('player2', AIConnection(game, 2, delay=0)))
    )

